import json
from decimal import ROUND_DOWN, Decimal
from pathlib import Path

PRECISION = Decimal(".01")


def main():
    project_root = Path(__file__).resolve().parent.parent
    coverage_summary = project_root / "coverage-summary.json"

    coverage_data = json.loads(coverage_summary.read_text(encoding="utf-8"))
    total_data = coverage_data.pop("total")

    lines = [
        "\n",
        "Package | Statements\n",
        "--- | ---\n",
    ]

    for package, data in sorted(coverage_data.items()):
        statements_covered = data["statements_covered"]
        statements = data["statements"]

        rate = Decimal(statements_covered) / Decimal(statements) * 100
        rate = rate.quantize(PRECISION, rounding=ROUND_DOWN)
        lines.append(
            f"{package} | {100 if rate == 100 else rate}% ({statements_covered} / {statements})\n"  # noqa: PLR2004
        )

    total_statements_covered = total_data["statements_covered"]
    total_statements = total_data["statements"]
    total_rate = Decimal(total_statements_covered) / Decimal(total_statements) * 100
    total_rate = total_rate.quantize(PRECISION, rounding=ROUND_DOWN)
    color = "ok" if float(total_rate) >= 95 else "critical"  # noqa: PLR2004
    lines.insert(0, f"![Code Coverage](https://img.shields.io/badge/coverage-{total_rate}%25-{color}?style=flat)\n")

    lines.append(
        f"**Summary** | {100 if total_rate == 100 else total_rate}% "  # noqa: PLR2004
        f"({total_statements_covered} / {total_statements})\n"
    )

    coverage_report = project_root / "coverage-report.md"
    with coverage_report.open("w", encoding="utf-8") as f:
        f.write("".join(lines))


if __name__ == "__main__":
    main()



# import polars as pl
#
# from renkon.core.derivation import Mean, Median
#
#
# def test_mean() -> None:
#     deriv = Mean()
#     assert deriv.derive_name("foo") == "mean(foo)"
#     assert deriv.derive_values(pl.Series([1, 2, 3, 4])).to_list() == [2.5] * 4
#
#
# def test_median() -> None:
#     deriv = Median()
#     assert deriv.derive_name("foo") == "median(foo)"
#     assert deriv.derive_values(pl.Series([1, 2, 3, 4])).to_list() == [2.5] * 4



from renkon.repo.registry import Registry, SQLiteRegistry
from renkon.repo.repository import Repository
from renkon.repo.storage import FileSystemStorage, Storage

__all__ = ["Repository", "Storage", "FileSystemStorage", "Registry", "SQLiteRegistry"]



from pathlib import Path

import pytest
from pyarrow import csv
from pyarrow import fs as pa_fs

from renkon.config import Config, load_config
from renkon.repo import Storage
from renkon.repo.registry import Registry, SQLiteRegistry
from renkon.repo.repository import Repository
from renkon.repo.storage import FileSystemStorage

TESTS_DIR = Path(__file__).parent

SEMICOLON_WITH_TYPE_ROW = {
    "parse_options": csv.ParseOptions(delimiter=";"),
    "read_options": csv.ReadOptions(skip_rows_after_names=1),
}

DEFAULT = {
    "parse_options": csv.ParseOptions(),
    "read_options": csv.ReadOptions(),
}

"""
List of sample datasets. Each key corresponds to a CSV file in the
`data` directory. Each contains the parse and read options needed
to read the file.
"""
SAMPLES = {
    "cars": SEMICOLON_WITH_TYPE_ROW,
    "cereals": SEMICOLON_WITH_TYPE_ROW,
    "cereals-corrupt": SEMICOLON_WITH_TYPE_ROW,
    "factbook": SEMICOLON_WITH_TYPE_ROW,
    "films": SEMICOLON_WITH_TYPE_ROW,
    "gini": DEFAULT,
    "smallwikipedia": SEMICOLON_WITH_TYPE_ROW,
}


@pytest.fixture(autouse=True)
def change_test_dir(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.chdir(tmp_path)


@pytest.fixture
def config(tmp_path: Path) -> Config:
    return load_config(repository={"path": tmp_path})


@pytest.fixture
def registry(config: Config) -> Registry:
    path = config.repository.path / "metadata.db"
    return SQLiteRegistry(path)


@pytest.fixture
def storage(config: Config) -> Storage:
    path = config.repository.path / "data"
    path.mkdir(parents=True, exist_ok=True)
    local_fs = pa_fs.LocalFileSystem(use_mmap=True)
    storage_fs = pa_fs.SubTreeFileSystem(str(path), local_fs)
    return FileSystemStorage(storage_fs)


@pytest.fixture
def repo(registry: Registry, storage: Storage) -> Repository:
    repo = Repository(registry=registry, storage=storage)
    for name, options in SAMPLES.items():
        data = csv.read_csv(TESTS_DIR / "samples" / f"{name}.csv", **options)
        repo.put(name, data)
    return repo



import atexit
import os
from pathlib import Path
from typing import Any, cast

import click
import pyarrow as pa
import pyarrow.fs as pa_fs
from loguru import logger
from pyarrow import csv as pa_csv
from pyarrow import ipc as pa_ipc
from pyarrow import parquet as pa_pq
from rich.logging import RichHandler

from renkon.__about__ import __version__
from renkon.client import RenkonFlightClient
from renkon.config import DEFAULTS, load_config
from renkon.repo import FileSystemStorage, Repository
from renkon.repo.registry import SQLiteRegistry
from renkon.server import RenkonFlightServer


def setup_default_logging() -> None:
    logger.configure(
        handlers=[
            {
                "sink": RichHandler(log_time_format=lambda dt: dt.isoformat(sep=" ", timespec="milliseconds")),
                "level": os.environ.get("LOG_LEVEL", "INFO"),
                "format": "{message}",
            }
        ]
    )


@click.group(
    context_settings={"help_option_names": ["-h", "--help"]},
    invoke_without_command=True,
)
@click.version_option(version=__version__, prog_name="renkon")
@click.pass_context
def cli(ctx: click.Context) -> None:
    """
    Starts a renkon server as a subprocess, and then connects a client to it.

    This is intended to be used only by an end-user for inspecting a repository,
    and is started with the default configuration. For real-world use, please
    use the server and client subcommands.

    If you must override the configuration, `renkon.toml` is loaded from the current working directory.

    By default, renkon runs on 127.0.0.1:1410, and uses the repository .renkon in the current working directory.
    """
    # If there is a subcommand, do nothing.
    if ctx.invoked_subcommand:
        return

    # No subcommand behavior.
    setup_default_logging()
    logger.critical("not yet implemented!")


@cli.command(context_settings={"show_default": True})
@click.argument("hostname", type=str, default=DEFAULTS["server"]["hostname"])
@click.argument("port", type=int, default=DEFAULTS["server"]["port"])
@click.option(
    "-d",
    "--data-dir",
    type=click.Path(path_type=Path),
    default=DEFAULTS["repository"]["path"],
    help="Path for data repository",
)
@click.pass_context
def server(_ctx: click.Context, hostname: str, port: int, data_dir: Path) -> None:
    # Configuration.
    config_overrides: dict[str, Any] = {
        "repository": {
            "path": data_dir.resolve(),
        },
        "server": {
            "hostname": hostname,
            "port": port,
        },
    }

    config = load_config(update_global=True, **config_overrides)

    # Logging.
    setup_default_logging()

    # Repository.
    repo_path = config.repository.path
    repo_path.mkdir(parents=True, exist_ok=True)

    fs = pa_fs.SubTreeFileSystem(str(repo_path / "data"), pa_fs.LocalFileSystem(use_mmap=True))

    repository = Repository(
        registry=SQLiteRegistry(repo_path / "metadata.db"),
        storage=FileSystemStorage(fs),
    )

    # Start server.
    logger.info(f"Starting Renkon server at {config.server.hostname}:{config.server.port}")
    server = RenkonFlightServer(repository, config.server)
    server.serve()


@cli.group(context_settings={"show_default": True})
@click.option("--hostname", "-H", type=str, default=DEFAULTS["server"]["hostname"])
@click.option("--port", "-P", type=int, default=DEFAULTS["server"]["port"])
@click.pass_context
def client(ctx: click.Context, hostname: str, port: int) -> None:
    # Logging.
    setup_default_logging()

    # Start client.
    client = RenkonFlightClient(location=f"grpc://{hostname}:{port}")

    logger.info(f"Connecting to {hostname}:{port}...")
    client.wait_for_available()
    logger.info("Connected!")

    ctx.obj = client

    atexit.register(client.close)

    # logger.info("Uploading test data (cereals-corrupt.csv)...")
    # table = csv.read_csv("etc/samples/cereals-corrupt.csv",
    #                      read_options=csv.ReadOptions(skip_rows_after_names=1),
    #                      parse_options=csv.ParseOptions(delimiter=";"),
    #                      convert_options=csv.ConvertOptions(auto_dict_encode=True))
    # client.upload("test", table)
    #
    # logger.info("Downloading test data...")
    # table = client.download("test")
    # logger.info(pretty_repr(table, expand_all=True))
    #
    # client.close()


@client.command(context_settings={"show_default": True})
@click.argument("name", type=str)
@click.argument("path", type=click.Path(path_type=Path, exists=True, dir_okay=False))
@click.pass_context
def put(ctx: click.Context, name: str, path: Path) -> None:
    client = cast(RenkonFlightClient, ctx.obj)

    table: pa.Table
    match path.suffix.lower():
        case ".csv":
            table = pa_csv.read_csv(path)
        case ".parquet":
            table = pa_pq.read_table(path)
        case ".arrow":
            table = pa_ipc.open_file(path).read_all()
        case _:
            msg = f"Unsupported file type: {path.suffix}"
            raise ValueError(msg)

    client.upload(name, table)
    logger.info(f"Uploaded {name}!")


@client.command(context_settings={"show_default": True})
@click.argument("name", type=str)
@click.argument("path", type=click.Path(path_type=Path, exists=False, dir_okay=False))
@click.pass_context
def get(ctx: click.Context, name: str, path: Path) -> None:
    client = cast(RenkonFlightClient, ctx.obj)

    if path.is_dir():
        logger.warning(f"Path {path} is a directory, defaulting to {path}/{name}.parquet")
        path.mkdir(parents=True, exist_ok=True)
        path = path / f"{name}.parquet"

    try:
        table = client.download(name)
        match path.suffix.lower():
            case ".csv":
                pa_csv.write_csv(table, path)
            case ".parquet":
                pa_pq.write_table(table, path)
            case ".arrow":
                pa_ipc.new_file(path, table.schema).write_all(table)
            case _:
                msg = f"Unsupported file type: {path.suffix}"
                raise ValueError(msg)
        logger.info(f"Downloaded {name} to {path}")
    except pa.ArrowKeyError:
        logger.error(f"Dataset {name} does not exist!")


if __name__ == "__main__":
    cli()



from collections import defaultdict
from collections.abc import Callable
from functools import partial
from multiprocessing import Event, Pool
from typing import Generic, TypeAlias, TypeVar

from loguru import logger

from renkon.task.result import Err, Ok, Result, Unk
from renkon.task.task import Task
from renkon.util.dag import DAG

_T = TypeVar("_T")

TaskSpec: TypeAlias = tuple[str, Callable[..., _T], list[str]]


class TaskGraph(Generic[_T]):
    """
    Implements a task graph with dependencies using python multiprocessing,
    generic in the return type of the tasks.

    The idea is that you instantiate one of these per frame of data, and it
    will run all the inference tasks in parallel, respecting dependencies.

    There is no special support for data sharing between tasks.
    """

    __slots__ = ("task_dag", "task_id_to_name", "task_id_to_result", "task_name_to_id")

    task_dag: DAG[Task[_T]]

    task_id_to_name: dict[int, str]
    task_id_to_result: dict[int, Result[_T]]
    task_name_to_id: dict[str, int]

    def __init__(self) -> None:
        self.task_dag = DAG[Task[_T]]()
        self.task_id_to_name = {}
        self.task_id_to_result = defaultdict(Unk)
        self.task_name_to_id = {}

    def add_task(self, name: str, func: Callable[..., _T], dependencies: list[int]) -> int:
        """
        Add a single task to the graph to be run after the dependencies.
        """
        if name in self.task_name_to_id:
            msg = f"Task {name} already exists!"
            raise ValueError(msg)

        # Wrap the task in a Task object.
        task = Task(name, func)
        task_id = self.task_dag.add_node(task, dependencies)

        # Store the mapping from task id to name (and back).
        self.task_id_to_name[task_id] = name
        self.task_name_to_id[name] = task_id

        return task_id

    def add_tasks(self, specs: list[TaskSpec[_T]]) -> dict[str, int]:
        """
        Add a batch of tasks to the graph. The task names are the keys of the
        dictionary, and the values are tuples of (name, func, dependencies).

        Dependencies should be expressed using task names.

        :returns: a mapping from task names to task ids.
        """
        # We return a mapping from task names to task ids.
        task_name_to_id = {}

        for name, func, dependency_names in specs:
            dep_ids = [self.task_name_to_id[name] for name in dependency_names]
            self.add_task(name, func, dep_ids)
            task_name_to_id[name] = self.task_name_to_id[name]

        return task_name_to_id

    def get_task(self, task_id: int) -> Task[_T]:
        return self.task_dag.get_node(task_id)

    def get_result(self, task_id: int) -> object:
        return self.task_id_to_result[task_id]

    def run(self) -> None:
        tasks_remaining = set(range(len(self.task_dag)))
        all_done = Event()

        logger.info("Scanning dependency tree...")
        tasks_next = {}
        for task_id in range(len(self.task_dag)):
            deps = self.task_dag.get_dependencies(task_id)
            tasks_next[task_id] = deps
            logger.info(f"{task_id} <- {deps}.")

        # todo: inject pool (allow for different pool implementations)
        with Pool() as pool:
            # pool = MultiprocessingPoolExecutor(_pool)

            def on_complete(task_id: int, result: _T) -> None:
                nonlocal tasks_remaining

                # Store the result
                self.task_id_to_result[task_id] = Ok(result)

                # First, check if we're all done.
                tasks_remaining.remove(task_id)
                if not tasks_remaining:
                    all_done.set()
                    return

                # Otherwise, submit the tasks that were only waiting for this one.
                for next_task_id in self.task_dag.get_dependents(task_id):
                    s = tasks_next[next_task_id]
                    s.remove(task_id)
                    if not s:
                        logger.info(f"All dependencies finished for {next_task_id}.")
                        submit(next_task_id)

            def on_error(task_id: int, error: Exception) -> None:
                nonlocal tasks_remaining

                logger.info(f"Task {task_id} failed: {error}")
                self.task_id_to_result[task_id] = Err(error)

                # Compute the set of tasks that depend on this one.
                pruned_tasks = self.task_dag.get_descendants(task_id)
                pruned_task_names = [self.get_task(task_id).name for task_id in pruned_tasks]
                logger.info(f"Pruning tasks: {pruned_task_names}")

                # Remove the failed task and all of its descendants.
                tasks_remaining.remove(task_id)
                tasks_remaining = tasks_remaining.difference(pruned_tasks)

                # Then, check if we're all done.
                if not tasks_remaining:
                    all_done.set()
                    return

            def submit(task_id: int) -> None:
                task = self.get_task(task_id)
                pool.apply_async(
                    task.func,
                    args=[task.name],
                    callback=partial(on_complete, task_id),
                    error_callback=partial(on_error, task_id),
                )

            # Kick off the tasks that have no dependencies.
            for task_id in self.task_dag.get_roots():
                submit(task_id)
            all_done.wait()


#
# def dummy_task(name: str) -> int:
#     proc = current_process()
#     logger.info(f"START {name}...")
#     time.sleep(0.1)
#     logger.info(f"DONE {name}...")
#     return proc.pid if proc.pid is not None else 0
#
#
# def fail_task(name: str) -> int:
#     current_process()
#     logger.info(f"START {name}...")
#     msg = "Uh oh, SNAFU!"
#     raise RuntimeError(msg)
#
#
# if __name__ == "__main__":
#     g = TaskGraph[int]()
#
#     name_to_tid = g.add_tasks(
#         [
#             ("Task-0", dummy_task, []),
#             ("Task-1", dummy_task, ["Task-0"]),
#             ("Task-2", dummy_task, []),
#             ("Task-3", dummy_task, ["Task-1", "Task-2"]),
#             ("Task-4", fail_task, ["Task-3"]),
#             ("Task-5", dummy_task, ["Task-4"]),
#             ("Task-6", dummy_task, ["Task-1", "Task-5"]),
#         ]
#     )
#
#     logger.info("Added tasks:\n" + pformat(name_to_tid))
#     g.run()
#
#     for name in sorted(name_to_tid.keys()):
#         task_id = name_to_tid[name]
#         logger.info(f"Task {name} has result {g.get_result(task_id)}.")



from ipaddress import IPv4Address
from pathlib import Path

import pytest

from renkon.config import load_config


def test_load_toml_config(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    # Monkeypatch the current working directory.
    monkeypatch.chdir(tmp_path)

    # Create a renkon.toml file.
    toml_path = tmp_path / "renkon.toml"
    toml_path.write_text(
        """
        [server]
        hostname = "1.2.3.4"
        port = 1234

        [repository]
        path = "foo/bar/baz"
        """
    )

    # Load the configuration.
    config = load_config()

    # Check that the configuration has been loaded correctly.
    assert config.server.hostname == IPv4Address("1.2.3.4")
    assert config.server.port == 1234
    assert config.repository.path.resolve() == tmp_path / "foo" / "bar" / "baz"



from __future__ import annotations

import sqlite3
from collections.abc import Callable
from dataclasses import dataclass
from pathlib import PurePath
from typing import Any, Literal, NamedTuple, Protocol, TypeAlias, TypeGuard, TypeVar

import pyarrow as pa

from renkon.util.common import deserialize_schema

FileType: TypeAlias = Literal["parquet", "arrow"]


# Type guard for FileType
def is_valid_filetype(value: str) -> TypeGuard[FileType]:
    return value in ("parquet", "arrow")


_RowT = TypeVar("_RowT")
_TupleT = TypeVar("_TupleT", bound=tuple[Any, ...])
RowFactory: TypeAlias = Callable[[sqlite3.Cursor, _TupleT], _RowT]


_T = TypeVar("_T")


class SupportsRowFactory(Protocol):
    @classmethod
    def row_factory(cls: type[_T], cur: sqlite3.Cursor, row: tuple[Any, ...]) -> _T:
        ...


class TableDBTuple(NamedTuple):
    """
    Tuple type for a table record in the metadata registry (database).
    """

    path: str
    name: str
    filetype: FileType
    schema: bytes
    rows: int
    size: int

    @classmethod
    def row_factory(cls: type[TableDBTuple], _cur: sqlite3.Cursor, row: tuple[Any, ...]) -> TableDBTuple:
        return cls(*row)


"""
Tuple type for a table record in the metadata registry (database).
"""


@dataclass(frozen=True, kw_only=True, slots=True)
class TableInfo:
    """
    Information record for a table according to the metadata registry (database).
    """

    path: PurePath
    name: str
    filetype: FileType
    schema: pa.Schema
    rows: int
    size: int

    @classmethod
    def from_tuple(cls, values: TableDBTuple) -> TableInfo:
        """
        Convert a tuple of values from the database into a TableInfo object.
        """
        if not is_valid_filetype(values.filetype):
            msg = f"Invalid filetype: {values.filetype}"
            raise ValueError(msg)
        return cls(
            name=values.name,
            path=PurePath(values.path),
            filetype=values.filetype,
            schema=deserialize_schema(values.schema),
            rows=values.rows,
            size=values.size,
        )

    @classmethod
    def from_stat(cls, name: str, stat: TableStat) -> TableInfo:
        """
        Create a TableInfo object from a TableStat object and name.
        """
        return cls(
            name=name,
            path=stat.path,
            filetype=stat.filetype,
            schema=stat.schema,
            rows=stat.rows,
            size=stat.size,
        )


@dataclass(frozen=True, kw_only=True, slots=True)
class TableStat:
    """
    Information about a table in storage, produced by the storage backend
    from the underlying data.

    :field path: Logical path to the table in storage.
    :field filetype: Filetype of the table, either "parquet" or "arrow".
    :field schema: Schema of the table.
    :field rows: Number of rows in the table, or -1 if unknown.
    :field size: Size of the serialized table in bytes, or -1 if unknown.
    """

    path: PurePath
    filetype: FileType
    schema: pa.Schema
    rows: int = -1
    size: int = -1



from __future__ import annotations

from collections.abc import Iterable
from typing import Generic, TypeVar

T_ = TypeVar("T_")


class DAG(Generic[T_]):
    """
    Generic append-only doubly-linked directed acyclic graph implementation.

    This is not a general purpose graph, it is specifically designed to
    be used in the Reactor for job scheduling. Therefore:

      - Adding edges is only allowed together with a new node.
        - Edges cannot be added between existing nodes.
      - Adding edges is only allowed if it does not create a cycle.
      - Removing nodes is not allowed.
      - Arrows are directed from dependencies to dependents.

    """

    _nodes: list[T_]
    _roots: list[int]
    _fwd_edges: dict[int, set[int]]
    _rev_edges: dict[int, set[int]]

    def __init__(self) -> None:
        self._nodes = []
        self._roots = []
        self._fwd_edges = {}
        self._rev_edges = {}

    def __len__(self) -> int:
        return len(self._nodes)

    def add_node(self, node: T_, dependencies: Iterable[int]) -> int:
        """
        Add a new node to the graph.

        :param node: The node to add.
        :param dependencies: The indices of the nodes that this node depends on.
        :return: The index of the new node.
        """
        node_index = len(self._nodes)
        self._nodes.append(node)
        self._fwd_edges[node_index] = set()
        self._rev_edges[node_index] = set()
        for dependency in dependencies:
            self._fwd_edges[dependency].add(node_index)
            self._rev_edges[node_index].add(dependency)
        if not dependencies:
            self._roots.append(node_index)
        return node_index

    def get_node(self, node_index: int) -> T_:
        """
        Get a node from the graph.

        :param node_index: The index of the node to get.
        :return: The node.
        """
        return self._nodes[node_index]

    def get_roots(self) -> set[int]:
        """
        Get the indices of the root nodes in the graph.

        :return: The indices of the root nodes.
        """
        return set(self._roots)

    def get_dependencies(self, node_index: int) -> set[int]:
        """
        Get the indices of the nodes that a node depends on.

        :param node_index: The index of the node.
        :return: The indices of the nodes that the node depends on.
        """
        return self._rev_edges[node_index]

    def get_dependents(self, node_index: int) -> set[int]:
        """
        Get the indices of the nodes that depend on a node.

        :param node_index: The index of the node.
        :return: The indices of the nodes that depend on the node.
        """
        return self._fwd_edges[node_index]

    def get_descendants(self, node_index: int) -> set[int]:
        """
        Get the indices of the nodes that are descendants of a node.

        :param node_index: The index of the node.
        :return: The indices of the nodes that are descendants of the node.
        """
        descendants = set()
        stack = [node_index]
        while stack:
            idx = stack.pop()
            descendants.add(idx)
            stack.extend(self._fwd_edges[idx])
        descendants.remove(node_index)
        return descendants



from dataclasses import dataclass
from typing import Generic, TypeAlias, TypeVar

_T = TypeVar("_T")


@dataclass(frozen=True, slots=True)
class Ok(Generic[_T]):
    """
    Represents a successful result, carrying the value.
    """

    value: _T

    def __repr__(self) -> str:
        return f"Ok({self.value!r})"


@dataclass(frozen=True, slots=True)
class Err:
    """
    Represents an error result, carrying the responsible exception.
    """

    cause: Exception

    def __repr__(self) -> str:
        return f"Err({self.cause!r})"


class Unk:
    """
    Represents an unknown (not yet computed?) result.
    """

    def __repr__(self) -> str:
        return "Unk"


Result: TypeAlias = Ok[_T] | Err | Unk



import json
from collections import defaultdict
from pathlib import Path

from lxml import etree  # nosec B410

PACKAGES = {
    "src/renkon/": "renkon",
    "tests/": "tests",
}


def main():
    project_root = Path(__file__).resolve().parent.parent
    coverage_report = project_root / "coverage.xml"
    root = etree.fromstring(coverage_report.read_text())  # nosec B320  # noqa: S320

    raw_package_data = defaultdict(lambda: {"hits": 0, "misses": 0})
    for package in root.find("packages"):
        for module in package.find("classes"):
            filename = module.attrib["filename"]
            for relative_path, package_name in PACKAGES.items():
                if filename.startswith(relative_path):
                    data = raw_package_data[package_name]
                    break
            else:
                message = f"unknown package: {module}"
                raise ValueError(message)

            for line in module.find("lines"):
                if line.attrib["hits"] == "1":
                    data["hits"] += 1
                else:
                    data["misses"] += 1

    total_statements_covered = 0
    total_statements = 0
    coverage_data = {}
    for package_name, data in sorted(raw_package_data.items()):
        statements_covered = data["hits"]
        statements = statements_covered + data["misses"]
        total_statements_covered += statements_covered
        total_statements += statements

        coverage_data[package_name] = {"statements_covered": statements_covered, "statements": statements}
    coverage_data["total"] = {"statements_covered": total_statements_covered, "statements": total_statements}

    coverage_summary = project_root / "coverage-summary.json"
    coverage_summary.write_text(json.dumps(coverage_data, indent=4), encoding="utf-8")


if __name__ == "__main__":
    main()



from __future__ import annotations

import re
from typing import NoReturn, cast

import pyarrow as pa


def unreachable() -> NoReturn:
    msg = "unreachable"
    raise AssertionError(msg)


SUP = {
    "0": "⁰",
    "1": "¹",
    "2": "²",
    "3": "³",
    "4": "⁴",
    "5": "⁵",
    "6": "⁶",
    "7": "⁷",
    "8": "⁸",
    "9": "⁹",
}

SUB = {
    "0": "₀",
    "1": "₁",
    "2": "₂",
    "3": "₃",
    "4": "₄",
    "5": "₅",
    "6": "₆",
    "7": "₇",
    "8": "₈",
    "9": "₉",
}

SUP_RE = re.compile(r"(?:\^|\^{)([0-9]+)}?")
SUB_RE = re.compile(r"(?:_|_{)([0-9]+)}?")


def pretty_sup(s: str) -> str:  # pragma: no cover
    """
    Format pretty superscripts in a string. For example:
      -  x^2 -> x²
      - x^{123} -> x¹²³
    """
    for match in SUP_RE.finditer(s):
        old = match.group(0)
        new = "".join(SUP.get(c, c) for c in match.group(1))
        s = s.replace(old, new)
    return s


def pretty_sub(s: str) -> str:  # pragma: no cover
    """
    Format pretty subscripts in a string. For example:
      -  x_2 -> x₂
      - x_{123} -> x₁₂₃
    """
    for match in SUB_RE.finditer(s):
        old = match.group(0)
        new = "".join(SUB.get(c, c) for c in match.group(1))
        s = s.replace(old, new)
    return s


def serialize_schema(schema: pa.Schema) -> bytes:
    """
    Serialize a schema to bytes.
    """
    return cast(bytes, schema.serialize().to_pybytes())


def deserialize_schema(blob: bytes) -> pa.Schema:
    """
    Deserialize a schema from bytes.
    """
    return pa.ipc.read_schema(pa.py_buffer(blob))



from pathlib import PurePath

import pyarrow as pa

from renkon.repo.info import TableInfo
from renkon.repo.registry import Registry
from renkon.repo.storage import Storage


class Repository:
    """
    The repository is the state of the system, containing all submitted inputs and
    computed outputs.

    Composes/mediates a Store and a Registry.

    The Store is responsible for storing and retrieving data on disk.
    The Registry is responsible for metadata queries, such as retrieving the path
    for a given input or output result.
    """

    _registry: Registry
    _storage: Storage

    def __init__(self, registry: Registry, storage: Storage) -> None:
        """
        Open a repository at the given path.
        """
        self._registry = registry
        self._storage = storage

    def get(self, name: str) -> pa.Table | None:
        if (table_info := self._registry.lookup(name, by="name")) is None:
            return None
        return self._storage.read(table_info.path)

    def put(self, name: str, table: pa.Table, *, for_ipc: bool = False, for_storage: bool = True) -> None:
        """
        Put data into the repository.
        """
        path = PurePath(name)
        if path.suffix:
            msg = f"Name '{name}' cannot have a suffix (extension)."
            raise ValueError(msg)

        paths = []
        if not for_ipc and not for_storage:
            msg = "Cannot store data for neither IPC nor storage."
            raise ValueError(msg)

        if for_ipc:
            paths.append(path.with_suffix(".arrow"))
        if for_storage:
            paths.append(path.with_suffix(".parquet"))

        for path in paths:
            self._storage.write(path, table)
            if (table_info := self._storage.info(path)) is None:
                msg = f"Table '{name}' not found in registry immediately after store. Something is broken."
                raise LookupError(msg)
            self._registry.register(name, path, table_info)

    def exists(self, name: str) -> bool:
        """
        Check if a table exists in the repository.
        """
        return self._registry.lookup(name, by="name") is not None

    def get_info(self, name: str) -> TableInfo | None:
        """
        Get information about a table in the repository.
        """
        return self._registry.lookup(name, by="name")

    def list_info(self) -> list[TableInfo]:
        """
        List all tables in the repository.
        """
        return self._registry.list_all()



from collections.abc import Callable
from typing import TypeVar

import pytest
from toolz import curry, identity

from renkon.task.graph import TaskGraph
from renkon.task.result import Err, Ok, Unk

T = TypeVar("T")


def identitwo(x: T, _: object) -> T:
    """Like identity, but with an extra argument that is ignored."""
    return x


def const(x: T) -> Callable[[object], T]:
    return curry(identitwo, x)  # type: ignore


def fail(_: T) -> T:
    msg = "fail"
    raise RuntimeError(msg)


def test_add_task() -> None:
    g: TaskGraph[int] = TaskGraph()
    id_a = g.add_task("a", curry(identity, 42), [])

    assert g.get_task(id_a).name == "a"
    assert g.get_task(id_a).func() == 42


def test_add_tasks() -> None:
    g: TaskGraph[int] = TaskGraph()

    g.add_tasks(
        [
            ("a", curry(identity, 42), []),
            ("b", curry(identity, 43), []),
        ]
    )


def test_no_duplicate_task_names() -> None:
    g: TaskGraph[int] = TaskGraph()
    g.add_task("a", curry(identity, 42), [])

    with pytest.raises(ValueError):
        g.add_task("a", curry(identity, 42), [])


def test_run_line() -> None:
    g: TaskGraph[int] = TaskGraph()

    g.add_tasks(
        [
            ("a", const(42), []),
            ("b", const(43), ["a"]),
            ("c", const(44), ["b"]),
        ]
    )

    g.run()

    assert g.get_result(g.task_name_to_id["a"]) == Ok(42)
    assert g.get_result(g.task_name_to_id["b"]) == Ok(43)
    assert g.get_result(g.task_name_to_id["c"]) == Ok(44)


def test_run_diamond() -> None:
    g: TaskGraph[int] = TaskGraph()

    g.add_tasks(
        [
            ("a", const(42), []),
            ("b", const(43), ["a"]),
            ("c", const(44), ["a"]),
            ("d", const(45), ["b", "c"]),
        ]
    )

    g.run()

    assert g.get_result(g.task_name_to_id["a"]) == Ok(42)
    assert g.get_result(g.task_name_to_id["b"]) == Ok(43)
    assert g.get_result(g.task_name_to_id["c"]) == Ok(44)
    assert g.get_result(g.task_name_to_id["d"]) == Ok(45)


def test_run_prune_line() -> None:
    g: TaskGraph[int] = TaskGraph()

    g.add_tasks(
        [
            ("a", const(42), []),
            ("b", fail, ["a"]),
            ("c", const(44), ["b"]),
        ]
    )

    g.run()

    assert type(g.get_result(g.task_name_to_id["a"])) is Ok
    assert type(g.get_result(g.task_name_to_id["b"])) is Err
    assert type(g.get_result(g.task_name_to_id["c"])) is Unk


def test_run_prune_complex() -> None:
    g: TaskGraph[int] = TaskGraph()

    g.add_tasks(
        [
            ("a", const(42), []),
            ("b", const(43), ["a"]),
            ("c", const(44), []),
            ("d", const(45), ["b", "c"]),
            ("e", fail, ["d"]),
            ("f", const(47), ["e"]),
            ("g", const(48), ["a", "f"]),
        ]
    )

    g.run()

    assert type(g.get_result(g.task_name_to_id["a"])) is Ok
    assert type(g.get_result(g.task_name_to_id["b"])) is Ok
    assert type(g.get_result(g.task_name_to_id["c"])) is Ok
    assert type(g.get_result(g.task_name_to_id["d"])) is Ok
    assert type(g.get_result(g.task_name_to_id["e"])) is Err
    assert type(g.get_result(g.task_name_to_id["f"])) is Unk
    assert type(g.get_result(g.task_name_to_id["g"])) is Unk



# import sys
# from pathlib import Path
#
# from loguru import logger
# from pyarrow import csv
# from rich.console import Console
#
# from renkon.repo import get_repo
#
# console = Console()
#
# SEMICOLON_WITH_TYPE_ROW = {
#     "parse_options": csv.ParseOptions(delimiter=";"),
#     "read_options": csv.ReadOptions(skip_rows_after_names=1),
# }
#
# DEFAULT = {
#     "parse_options": csv.ParseOptions(),
#     "read_options": csv.ReadOptions(),
# }
#
# SAMPLES = {
#     "cars": SEMICOLON_WITH_TYPE_ROW,
#     "cereals": SEMICOLON_WITH_TYPE_ROW,
#     "cereals-corrupt": SEMICOLON_WITH_TYPE_ROW,
#     "factbook": SEMICOLON_WITH_TYPE_ROW,
#     "films": SEMICOLON_WITH_TYPE_ROW,
#     "gini": DEFAULT,
#     "smallwikipedia": SEMICOLON_WITH_TYPE_ROW,
# }
#
# # SKETCHES = []
#
# repo = get_repo()
# for name, options in SAMPLES.items():
#     data = csv.read_csv(Path.cwd() / "etc/samples" / f"{name}.csv", **options)
#     repo.put_input_table(name, data)
#     logger.info(f"Loaded sample {name} into the repository.")
#
# if __name__ == "__main__":
#     from renkon.cli import cli
#
#     for name in ["cereals"]:
#         table = repo.get_input_dataframe(name)
#         console.print(f"[bold]{name}[/bold]")
#         console.print(table)
#         console.print()
#
#     sys.exit(cli())



from abc import abstractmethod
from pathlib import PurePath
from typing import Protocol, TypeAlias, cast

import pyarrow as pa
import pyarrow.fs as pa_fs
import pyarrow.ipc as pa_ipc
import pyarrow.parquet as pa_pq

from renkon.repo.info import TableStat

StoragePath: TypeAlias = PurePath


class Storage(Protocol):  # pragma: no cover
    """
    Protocol for a storage backend for a :class:`renkon.repo.Repo`.

    A storage backend is responsible for storing and retrieving data from e.g. disk. It is not
    concerned with metadata, nor the contents of the data itself. It is however concerned with
    the format of the data, and how it is stored on disk (or elsewhere).

    The storage backend furthermore is not aware of the choice of dataframe library (Polars, Python, etc)
    and deals strictly in Arrow Tables.

    The purpose of this abstraction is to allow for different storage backends to be used, and to
    abstract storage-level details such as storage format away from usage-level abstractions such as
    intended usage (processing, storage, etc).
    """

    @abstractmethod
    def read(self, path: StoragePath) -> pa.Table | None:
        """
        Return a Table from the storage backend, or None if it does not exist.
        """
        ...

    def write(self, path: StoragePath, table: pa.Table) -> None:
        """
        Put a Table into the storage backend. Overwrites any existing data at the given path.
        """
        ...

    def delete(self, path: StoragePath) -> None:
        """
        Delete the data at the given path from the storage backend.
        """
        ...

    def info(self, path: StoragePath) -> TableStat | None:
        """
        Return a StoredTableInfo with metadata about the given table, such as
        size in bytes, number of records, etc. Useful for flights.
        """
        ...

    def exists(self, path: StoragePath) -> bool:
        """
        Return True if the path exists in the storage backend, False otherwise.
        """
        ...


class FileSystemStorage(Storage):
    """
    Abstracts details of disk storage of data away from the Repo.

    Recommended to initialize with a FileSystem with memory-mapping enabled,
    such as pyarrow.fs.LocalFileSystem(use_mmap=True), to avoid unnecessary
    memory copies (if the data is already memory-mapped by another process,
    and the OS optimizes for this).
    """

    fs: pa_fs.FileSystem

    def __init__(self, fs: pa_fs.FileSystem) -> None:
        self.fs = fs

    def read(self, path: StoragePath) -> pa.Table | None:
        match path.suffix:
            case ".parquet":
                return pa_pq.read_table(str(path), filesystem=self.fs)
            case ".arrow":
                with self.fs.open_input_file(str(path)) as file:
                    reader = pa_ipc.RecordBatchStreamReader(file)
                    table = reader.read_all()
                return table
            case _:
                msg = f"Unknown file extension: {path.suffix}"
                raise ValueError(msg)

    def write(self, path: StoragePath, table: pa.Table) -> None:
        self.fs.create_dir(str(path.parent), recursive=True)
        match path.suffix:
            case ".parquet":
                pa_pq.write_table(table, str(path), filesystem=self.fs)
            case ".arrow":
                with self.fs.open_output_stream(str(path)) as stream:
                    writer = pa_ipc.RecordBatchStreamWriter(stream, table.schema)
                    writer.write(table)
                    writer.close()
            case _:
                msg = f"Unknown file extension: {path.suffix}"
                raise ValueError(msg)

    def delete(self, path: StoragePath) -> None:
        self.fs.delete_file(str(path))

    def info(self, path: StoragePath) -> TableStat | None:
        match path.suffix:
            case ".parquet":
                schema = pa_pq.read_schema(str(path), filesystem=self.fs)
                metadata = pa_pq.read_metadata(str(path), filesystem=self.fs)
                file_info = self.fs.get_file_info(str(path))
                return TableStat(
                    path=path,
                    filetype="parquet",
                    schema=schema,
                    rows=metadata.num_rows,
                    size=file_info.size,
                )
            case ".arrow":
                with self.fs.open_input_file(str(path)) as file:
                    reader = pa_ipc.RecordBatchStreamReader(file)
                    table = reader.read_all()  # todo: ensure this plays well with use_mmap

                    return TableStat(
                        path=path,
                        filetype="arrow",
                        schema=table.schema,
                        rows=table.num_rows,
                        size=table.nbytes,
                    )
            case _:
                msg = f"Unknown file extension: {path.suffix}"
                raise ValueError(msg)

    def exists(self, path: StoragePath) -> bool:
        file_info: pa_fs.FileInfo = self.fs.get_file_info(str(path))
        return cast(bool, file_info.is_file)



from __future__ import annotations

import sqlite3
from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path, PurePath
from sqlite3 import Connection as SQLiteConnection
from typing import Literal, Protocol, TypeAlias

from renkon.repo.info import SupportsRowFactory, TableDBTuple, TableInfo, TableStat
from renkon.repo.queries import queries
from renkon.util.common import serialize_schema

RegistryLookupKey: TypeAlias = Literal["name", "path"]
RegistrySearchKey: TypeAlias = Literal["name", "path"]


class Registry(Protocol):  # pragma: no cover
    def register(self, name: str, path: PurePath, table_info: TableStat) -> None:
        ...

    def unregister(self, name: str) -> None:
        ...

    def list_all(self) -> list[TableInfo]:
        ...

    def lookup(self, key: str, *, by: RegistryLookupKey) -> TableInfo | None:
        ...

    def search(self, query: str = "*", *, by: RegistrySearchKey) -> list[TableInfo]:
        ...


class SQLiteRegistry(Registry):
    """
    Handles all things related to metadata, composed by Repo.
    You should generally not need to interact with this class directly.
    """

    db_path: Path | Literal[":memory:"]

    def __init__(self, db_path: Path) -> None:
        self.db_path = db_path
        self._create_tables()

    @contextmanager
    def _connect(self, *, row_type: type[SupportsRowFactory] | None = None) -> Generator[SQLiteConnection, None, None]:
        """
        Get a connection to the metadata repository. This must be used in each method
        to avoid persisting a connection and risk it being used by multiple threads.

        :param row_type: a type which provides a row_factory method to set on the connection.
        """
        try:
            # note: using a connection object as a context manager implies
            # that commit will be called if no exception is raised, and rollback
            # otherwise.
            with sqlite3.connect(self.db_path) as conn:
                if row_type is not None:
                    conn.row_factory = row_type.row_factory
                yield conn
        finally:
            conn.close()

    def _create_tables(self) -> None:
        """
        Create tables in the metadata repository.
        """
        with self._connect() as conn:
            queries.create_tables(conn)

    def register(self, name: str, path: PurePath, table_info: TableStat) -> None:
        """
        Register a table.
        """
        with self._connect() as conn:
            queries.register_table(
                conn,
                path=str(path),
                name=name,
                filetype=table_info.filetype,
                schema=serialize_schema(table_info.schema),
                rows=table_info.rows,
                size=table_info.size,
            )

    def unregister(self, name: str) -> None:
        """
        Unregister a table.
        """
        with self._connect() as conn:
            queries.unregister_table(conn, name=name)

    def list_all(self) -> list[TableInfo]:
        """
        List all tables.
        """
        with self._connect(row_type=TableDBTuple) as conn:
            row_tuples = queries.list_tables(conn)
            return [TableInfo.from_tuple(row_tuple) for row_tuple in row_tuples]

    def lookup(self, key: str, *, by: RegistryLookupKey) -> TableInfo | None:
        row_tuple = None

        with self._connect(row_type=TableDBTuple) as conn:
            match by:
                case "name":
                    # Prefer parquet to arrow if both exist.
                    row_tuple = queries.get_table(conn, name=key, filetype="parquet")
                    row_tuple = row_tuple or queries.get_table(conn, name=key, filetype="arrow")
                case "path":
                    # The string path stored in the database is native (e.g. on Win: "foo/bar" -> "foo\\bar").
                    native_path = str(Path(key))
                    row_tuple = queries.get_table_by_path(conn, path=native_path)

        if row_tuple is None:
            return None

        return TableInfo.from_tuple(row_tuple)

    def search(self, query: str = "*", *, by: RegistrySearchKey) -> list[TableInfo]:
        row_tuples: list[TableDBTuple] = []
        with self._connect(row_type=TableDBTuple) as conn:
            match by:
                case "name":
                    # Prefer parquet to arrow if both exist
                    row_tuples = queries.search_tables_by_name(conn, name=query)
                case "path":
                    row_tuples = queries.search_tables_by_path(conn, path=query)

        return [TableInfo.from_tuple(values) for values in row_tuples]



from collections.abc import Callable
from dataclasses import dataclass
from typing import Generic, TypeVar

_T = TypeVar("_T")


@dataclass(frozen=True, slots=True)
class Task(Generic[_T]):
    """
    Generic task class, with a name and a function to run.
    """

    name: str
    func: Callable[..., _T]



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Generic, Protocol, TypeVar

from polars import DataFrame, Series

if TYPE_CHECKING:
    from renkon.core.inference import InferenceStrategy

_TraitT = TypeVar("_TraitT", bound="Trait")
_PropT = TypeVar("_PropT", bound="PropTrait")
_StatT = TypeVar("_StatT", bound="StatTrait")


class InferenceStrategyKind(Enum):
    """Enumeration for inference strategies."""

    SIMPLE = 0
    """Simple inference strategy. Evaluated by simple predicate."""

    RANSAC = 1
    """RANSAC inference strategy. Valid only for statistical traits."""

    IQR = 2
    """1.5xIQR inference strategy. Valid only for a weaker prior of normality, more robust for smaller datasets."""

    THREE_SIGMA = 3
    """Three-sigma inference strategy. Valid only with a strong prior of normality and sufficient data."""


@dataclass(eq=True, frozen=True, kw_only=True, slots=True)
class TraitSketch(Generic[_TraitT]):
    trait_type: type[_TraitT]
    column_ids: tuple[str, ...]


class Trait(Protocol):
    @classmethod
    @abstractmethod
    def sketch(cls: type[_TraitT], column_ids: list[str]) -> TraitSketch[_TraitT]:
        """
        :return: a hashable token that uniquely identifies a sketch given some column IDs.
        """
        ...

    @classmethod
    @abstractmethod
    def inference_strategy(cls, priors: tuple[TraitSketch[Trait], ...]) -> InferenceStrategy:
        """
        :return: the inference strategy used by this invariant.

        :note: the inference strategy chosen may vary based on provided priors.
        """
        ...

    @classmethod
    @abstractmethod
    def arities(cls) -> tuple[int, ...]:
        """
        :return: the arities supported by this invariant.
        """
        ...

    @classmethod
    @abstractmethod
    def commutors(cls, arity: int) -> tuple[bool, ...]:
        """
        :return: whether this invariant is commutative for each position up to the given arity.

        :note: each position marked True can be swapped with any other position marked True without
               the invariant being considered distinct. For example: Equality.commutors(2) == [True, True]
        """
        ...

    @property
    @abstractmethod
    def data(self) -> DataFrame:
        """
        :return: the data this trait was inferred from.
        """
        ...

    @abstractmethod
    def test(self, data: DataFrame) -> Series:
        """
        :return: boolean Series of whether the trait holds on the given data (for each row).
        """
        ...


class TypeTrait(Trait, ABC):
    """
    A trait representing a type, e.g. "x is a string" or "x is a number".

    :note: These are in general not inferred but rather provided by the user or taken from the schema.
    """

    # note: should there be a test_coercible?

    pass


class PropTrait(Trait, ABC):
    """
    A trait representing a logical proposition, e.g. "x != 0" or "x < y".
    """

    @classmethod
    def satisfy(cls: type[_PropT], data: DataFrame) -> _PropT | None:
        """
        Attempts to find an assignment of variables (model) that satisfies the proposition on the given data,
        returning a trait instance if successful, or None if no such assignment can be found.
        """
        raise NotImplementedError

    def test(self, data: DataFrame) -> Series:
        return self.test_satisfied(data)

    def test_satisfied(self, data: DataFrame) -> Series:
        """
        :return: boolean Series of whether the proposition holds on the given data (for each row).
        """
        raise NotImplementedError


class StatTrait(Trait, ABC):
    """
    A trait representing a statistical property, e.g. "x is normally distributed" or "x is linearly correlated with y".
    """

    @classmethod
    def fit(cls: type[_StatT], data: DataFrame) -> _StatT | None:
        """
        Attempts to fit the statistical property to the given data, returning a trait instance if successful, or
        None if sufficient confidence/goodness-of-fit cannot be achieved.
        """
        raise NotImplementedError

    def test(self, data: DataFrame) -> Series:
        return self.test_inlying(data)

    def test_inlying(self, data: DataFrame) -> Series:
        """
        Tests whether each row of the given data is inlying with respect to the statistical property.

        ----

        *Examples:*

        For a normal distribution, this could be whether the data is within 3 standard deviations of the mean.

        For a linear correlation, this could be whether the data is within a 95% confidence interval.

        :return: boolean series of whether the data is inlying with respect to the statistical property (for each row).
        """
        raise NotImplementedError



from renkon.task.result import Err, Ok, Unk


def test_result_ok() -> None:
    ok = Ok(42)
    assert ok.value == 42
    assert repr(ok) == "Ok(42)"


def test_result_err() -> None:
    err = Err(ValueError("oops"))
    assert err.cause.args == ("oops",)
    assert repr(err) == "Err(ValueError('oops'))"


def test_result_unk() -> None:
    unk = Unk()
    assert repr(unk) == "Unk"



from renkon.util.dag import DAG


def test_dag_line() -> None:
    d = DAG[str]()
    id_a = d.add_node("a", [])
    id_b = d.add_node("b", [id_a])
    id_c = d.add_node("c", [id_b])
    assert d.get_dependents(id_a) == {id_b}
    assert d.get_dependents(id_b) == {id_c}
    assert d.get_dependents(id_c) == set()


def test_dag_disjoint() -> None:
    d = DAG[str]()
    id_a = d.add_node("a", [])
    id_b = d.add_node("b", [id_a])
    id_c = d.add_node("c", [id_b])

    id_x = d.add_node("x", [])
    id_y = d.add_node("y", [id_x])
    id_z = d.add_node("z", [id_y])

    assert d.get_roots() == {id_a, id_x}

    assert d.get_descendants(id_a) == {id_b, id_c}
    assert d.get_descendants(id_x) == {id_y, id_z}


def test_dag_diamond() -> None:
    d = DAG[str]()
    id_a = d.add_node("a", [])
    id_b = d.add_node("b", [id_a])
    id_c = d.add_node("c", [id_a])
    id_d = d.add_node("d", [id_b, id_c])

    assert d.get_roots() == {id_a}

    assert d.get_descendants(id_a) == {id_b, id_c, id_d}
    assert d.get_descendants(id_b) == {id_d}
    assert d.get_descendants(id_c) == {id_d}
    assert d.get_descendants(id_d) == set()



import pyarrow as pa
import pytest

from renkon.repo import Registry
from renkon.repo.repository import Repository

DATA = pa.Table.from_pydict(
    mapping={
        "a": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        "b": ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j"],
        "c": [True, False, True, False, True, False, True, False, True, False],
    }
)


def test_round_trip(repo: Repository) -> None:
    repo.put("foo/bar", DATA)
    assert repo.exists("foo/bar")
    data = repo.get("foo/bar")
    assert data is not None
    assert data.num_rows == 10
    assert data.num_columns == 3
    assert data.column_names == ["a", "b", "c"]
    assert data.equals(DATA)


def test_put_for_storage_only(repo: Repository, registry: Registry) -> None:
    repo.put("foo/bar", DATA, for_storage=True, for_ipc=False)
    assert registry.lookup("foo/bar", by="name") is not None
    assert registry.lookup("foo/bar.parquet", by="path") is not None


def test_put_for_ipc_only(repo: Repository, registry: Registry) -> None:
    repo.put("foo/bar", DATA, for_ipc=True, for_storage=True)
    assert registry.lookup("foo/bar", by="name") is not None
    assert registry.lookup("foo/bar.arrow", by="path") is not None


def test_put_for_both(repo: Repository, registry: Registry) -> None:
    repo.put("foo/bar", DATA, for_ipc=True, for_storage=True)
    assert registry.lookup("foo/bar", by="name") is not None
    assert registry.lookup("foo/bar.arrow", by="path") is not None
    assert registry.lookup("foo/bar.parquet", by="path") is not None


def test_put_for_neither_fails(repo: Repository, registry: Registry) -> None:
    with pytest.raises(ValueError, match="Cannot store data for neither IPC nor storage."):
        repo.put("foo/bar", DATA, for_ipc=False, for_storage=False)
    assert registry.lookup("foo/bar", by="name") is None
    assert registry.lookup("foo/bar.parquet", by="path") is None
    assert registry.lookup("foo/bar.arrow", by="path") is None



# """
# Derived columns are columns that are derived from other columns, but exist only within
# Renkon. This allows for inference over values derived from columns that are not present
# in the input table.
#
# For example, if the input table has columns `x` and `y`, it may be desirable to infer a
# trait such as `x ≥ median(y)`.
# """
# from collections.abc import Callable
# from functools import partial
# from typing import Generic, Protocol, TypeVar
#
# import polars as pl
# from polars import Series
# from polars.type_aliases import PythonLiteral
#
#
# class Derivation(Protocol):
#     name: str
#
#     def derive_name(self, column_name: str) -> str:
#         """
#         Produce a name for the derived column based on the name of the input column.
#         """
#         return f"{self.name}({column_name})"
#
#     def derive_values(self, column_values: Series) -> Series:
#         """
#         Produce an expression for the derived column based on the name of the input column.
#
#         In the case that the derived column is a constant, the result will contain that constant
#         repeated for each row in the input column.
#         """
#         ...
#
#
# _T = TypeVar("_T", bound=PythonLiteral)
#
#
# class Next(Derivation):
#     """
#     The Next derivation maps each element of the column to its successor.
#     """
#
#     name = "next"
#
#     def derive_values(self, column_values: Series) -> Series:
#         return column_values.shift(-1)
#
#
# class AggregateFunction(Derivation, Generic[_T]):
#     """
#     An aggregate derivation is a derivation that can be computed using an aggregate function
#     over the input column. This class exists mostly to cut down on boilerplate.
#     """
#
#     name: str
#     aggregate_fn: Callable[[Series], _T]
#
#     def derive_name(self, column_name: str) -> str:
#         return f"{self.name}({column_name})"
#
#     def derive_values(self, column_values: Series) -> Series:
#         return pl.repeat(value=self.aggregate_fn(column_values), n=column_values.len(), eager=True)
#
#
# class Mean(AggregateFunction[float]):
#     name = "mean"
#     aggregate_fn = partial(pl.mean)
#
#
# class Median(AggregateFunction[float]):
#     name = "median"
#     aggregate_fn = partial(pl.median)



import pyarrow as pa
from pyarrow.flight import FlightClient


class RenkonFlightClient(FlightClient):  # type: ignore[misc]
    def upload(self, name: str, table: pa.Table) -> None:
        descriptor = pa.flight.FlightDescriptor.for_path(name)
        writer, _ = self.do_put(descriptor, table.schema)
        writer.write(table)
        writer.close()

    def download(self, name: str) -> pa.Table:
        descriptor = pa.flight.FlightDescriptor.for_path(name)
        flight = self.get_flight_info(descriptor)
        reader = self.do_get(flight.endpoints[0].ticket)
        return reader.read_all()



from __future__ import annotations

from dataclasses import dataclass
from typing import Protocol

from polars import DataFrame, Series

from renkon.core.trait import Trait, TraitSketch, _TraitT


class InferenceStrategy(Protocol):
    def infer(self, sketch: TraitSketch[_TraitT], data: DataFrame) -> _TraitT:
        """
        :return: an invariant inferred from the given data.
        """
        raise NotImplementedError

    def test(self, trait: Trait, data: DataFrame) -> Series:
        """
        :param trait: the trait to evaluate.
        :param data: the data to evaluate on (may include the training data, but also test data)
        :return: a boolean series indicating the rows where the trait holds.
        """
        raise NotImplementedError

    def confidence(self, trait: Trait, data: DataFrame) -> float:
        """
        :param trait: the trait to evaluate.
        :param data: the data to evaluate on (may include the training data, but also test data)
        :return: evaluate the confidence of the inferred trait on the given data.
        """
        raise NotImplementedError


@dataclass(frozen=True, kw_only=True, slots=True)
class RANSACInferenceStrategy(InferenceStrategy):
    min_sample: int
    max_iterations: int = 3
    min_inlier_ratio: float = 0.90
    min_confidence: float = 0.90

    def infer(self, sketch: TraitSketch[_TraitT], data: DataFrame) -> _TraitT:
        raise NotImplementedError  # todo: implement

    def test(self, trait: Trait, data: DataFrame) -> Series:
        raise NotImplementedError  # todo: implement

    def confidence(self, trait: Trait, data: DataFrame) -> float:
        raise NotImplementedError  # todo: implement


@dataclass(frozen=True, kw_only=True, slots=True)
class SimpleInferenceStrategy(InferenceStrategy):
    """
    Simple inference strategy which evaluates a simple predicate on the data.

    By default, this evaluates on _all_ data points.
    TODO: adjust reported confidence for < 100% sample ratio.
    """

    sample_ratio: float = 1.0

    def infer(self, sketch: TraitSketch[_TraitT], data: DataFrame) -> _TraitT:
        raise NotImplementedError  # todo: implement

    def test(self, trait: Trait, data: DataFrame) -> Series:
        raise NotImplementedError  # todo: implement

    def confidence(self, trait: Trait, data: DataFrame) -> float:
        raise NotImplementedError  # todo: implement


@dataclass(frozen=True, kw_only=True, slots=True)
class ThreeSigmaInferenceStrategy(InferenceStrategy):
    def infer(self, sketch: TraitSketch[_TraitT], data: DataFrame) -> _TraitT:
        raise NotImplementedError  # todo: implement

    def test(self, trait: Trait, data: DataFrame) -> Series:
        raise NotImplementedError

    def confidence(self, trait: Trait, data: DataFrame) -> float:
        raise NotImplementedError



from collections.abc import Generator
from typing import Any

from loguru import logger
from pyarrow.flight import (
    Action,
    FlightDataStream,
    FlightDescriptor,
    FlightEndpoint,
    FlightInfo,
    FlightMetadataWriter,
    FlightServerBase,
    MetadataRecordBatchReader,
    RecordBatchStream,
    ServerCallContext,
    Ticket,
)

from renkon.config import ServerConfig
from renkon.repo import Repository
from renkon.repo.info import TableInfo


class RenkonFlightServer(FlightServerBase):  # type: ignore[misc]
    _location: str
    _repo: Repository
    _config: ServerConfig

    def __init__(self, repo: Repository, config: ServerConfig, **kwargs: Any):
        location = f"grpc://{config.hostname}:{config.port}"
        super().__init__(location, **kwargs)
        self._location = location
        self._repo = repo
        self._config = config

    def _make_flight_info(self, table_info: TableInfo) -> FlightInfo:
        name = table_info.name
        descriptor = FlightDescriptor.for_path(name.encode("utf-8"))
        endpoints = [FlightEndpoint(name, [self._location])]
        return FlightInfo(table_info.schema, descriptor, endpoints, table_info.rows, table_info.size)

    def list_flights(self, _context: ServerCallContext, _criteria: bytes) -> Generator[FlightInfo, None, None]:
        for table_info in self._repo.list_info():
            yield self._make_flight_info(table_info)

    def get_flight_info(self, _context: ServerCallContext, descriptor: FlightDescriptor) -> FlightInfo:
        name = descriptor.path[0].decode("utf-8")
        table_info = self._repo.get_info(name)
        if table_info is None:
            msg = f"Table {name} not found."
            raise KeyError(msg)
        return self._make_flight_info(table_info)

    def do_put(
        self,
        _context: ServerCallContext,
        descriptor: FlightDescriptor,
        reader: MetadataRecordBatchReader,
        _writer: FlightMetadataWriter,
    ) -> None:
        name = descriptor.path[0].decode("utf-8")
        table = reader.read_all()
        self._repo.put(name, table)

    def do_get(
        self,
        _context: ServerCallContext,
        ticket: Ticket,
    ) -> FlightDataStream:
        name = ticket.ticket.decode("utf-8")
        table = self._repo.get(name)
        return RecordBatchStream(table)

    def list_actions(self, _context: ServerCallContext) -> list[tuple[str, str]]:
        return [("dummy", "Dummy action, does nothing.")]

    def do_action(self, _context: ServerCallContext, action: Action) -> None:
        if action.type == "dummy":
            logger.info("Dummy action received, doing nothing.")
            pass
        else:
            msg = f"Action type {action.type} is not supported."
            raise NotImplementedError(msg)



from __future__ import annotations

import tomllib
from dataclasses import dataclass, field
from ipaddress import IPv4Address
from pathlib import Path
from typing import Any

from dacite import Config as DaciteConfig
from dacite import from_dict

# todo: load this from a default renkon.toml
DEFAULTS: dict[str, dict[str, Any]] = {
    "repository": {
        "path": Path(".renkon"),
    },
    "server": {
        "hostname": IPv4Address("127.0.0.1"),
        "port": 1410,  # stroke counts of 蓮根 (renkon)
    },
}


@dataclass(frozen=True, kw_only=True, slots=True)
class RepositoryConfig:
    """
    Renkon repository configuration class.
    """

    path: Path = field(default=DEFAULTS["repository"]["path"])


@dataclass(frozen=True, kw_only=True, slots=True)
class ServerConfig:
    """
    Renkon server configuration class.
    """

    hostname: IPv4Address = field(default=DEFAULTS["server"]["hostname"])
    port: int = field(default=DEFAULTS["server"]["port"])


@dataclass(frozen=True, kw_only=True, slots=True)
class Config:
    """
    Renkon configuration class.
    """

    repository: RepositoryConfig = field(default_factory=RepositoryConfig)
    server: ServerConfig = field(default_factory=ServerConfig)


def load_config(*, update_global: bool = False, **overrides: Any) -> Config:
    """
    Load the configuration from renkon.toml, and apply overrides.

    :param update_global: Whether to update the global configuration.
    :param overrides: Configuration overrides.
    :return: The configuration.
    """

    # If renkon.toml exists in the working directory, load it using pathlib.
    conf_data = {}
    if (p := Path.cwd() / "renkon.toml").exists():
        with p.open("rb") as f:
            conf_data = tomllib.load(f)

    # Apply overrides.
    conf_data.update(overrides)

    conf = from_dict(data_class=Config, data=conf_data, config=DaciteConfig(cast=[Path, IPv4Address]))

    if update_global:
        global config  # noqa: PLW0603
        config = conf

    return conf


# Global configuration. Should not change after the first load,
# except for testing-specific purposes, or for CLI overrides.
config = load_config()



from pathlib import Path
from sqlite3 import Connection as SQLiteConnection
from typing import Protocol, cast

import aiosql.queries

from renkon.repo.info import FileType, TableDBTuple


class RegistryQueries(Protocol):  # pragma: no cover
    """
    Protocol for the queries used by the registry. This allows us to define a typed interface
    for the queries supported by the registry, and also to support mocking use of the queries in
    tests without any actual database.

    Warning: if the types in this class are changed, the corresponding types in registry.sql must
    also be changed.
    """

    def create_tables(self, conn: SQLiteConnection) -> None:
        ...

    def register_table(
        self, conn: SQLiteConnection, *, path: str, name: str, filetype: str, schema: bytes, rows: int, size: int
    ) -> None:
        ...

    def unregister_table(self, conn: SQLiteConnection, *, name: str) -> None:
        ...

    def get_table(self, conn: SQLiteConnection, *, name: str, filetype: FileType) -> TableDBTuple:
        ...

    def get_table_by_path(self, conn: SQLiteConnection, *, path: str) -> TableDBTuple:
        ...

    def list_tables(self, conn: SQLiteConnection) -> list[TableDBTuple]:
        ...

    def search_tables_by_path(self, conn: SQLiteConnection, *, path: str) -> list[TableDBTuple]:
        ...

    def search_tables_by_name(self, conn: SQLiteConnection, *, name: str) -> list[TableDBTuple]:
        ...


class Queries(aiosql.queries.Queries, RegistryQueries):  # type: ignore[misc] # pragma: no cover
    """
    Queries for the repository.

    These methods do not define the logic for the queries, but rather the interface for the
    queries. This is in order to provide type hints for the queries. The actual logic is
    generated by AIOSQL at runtime. The queries are defined in repository.sql.
    """


queries: Queries = cast(
    Queries, aiosql.from_path(sql_path=Path(__file__).with_name("registry.sql"), driver_adapter="sqlite3")
)



from pathlib import Path, PurePath

import pyarrow as pa
import pytest

from renkon.repo import Storage

TABLE = pa.Table.from_pydict(
    mapping={"a": [1, 2, 3, 4, 5], "b": ["a", "b", "c", "d", "e"], "c": [True, False, True, False, True]}
)


def test_write_read_delete_parquet(storage: Storage) -> None:
    path = PurePath("foo/bar.parquet")

    storage.write(path, TABLE)
    assert storage.exists(path)

    table = storage.read(path)
    assert table is not None
    assert table.num_rows == 5
    assert table.num_columns == 3
    assert table.column_names == ["a", "b", "c"]
    assert table.column("a").to_pylist() == [1, 2, 3, 4, 5]
    assert table.column("b").to_pylist() == ["a", "b", "c", "d", "e"]
    assert table.column("c").to_pylist() == [True, False, True, False, True]

    storage.delete(path)
    assert not storage.exists(path)


def test_write_read_delete_arrow(storage: Storage) -> None:
    path = PurePath("foo/bar.parquet")

    storage.write(path, TABLE)
    assert storage.exists(path)

    table = storage.read(path)
    assert table is not None
    assert table.num_rows == 5
    assert table.num_columns == 3
    assert table.column_names == ["a", "b", "c"]
    assert table.column("a").to_pylist() == [1, 2, 3, 4, 5]
    assert table.column("b").to_pylist() == ["a", "b", "c", "d", "e"]
    assert table.column("c").to_pylist() == [True, False, True, False, True]

    storage.delete(path)
    assert not storage.exists(path)


def test_info_parquet(storage: Storage) -> None:
    path = PurePath("foo/bar.parquet")

    storage.write(path, TABLE)
    assert storage.exists(path)

    info = storage.info(path)
    assert info is not None
    assert info.path == path
    assert info.filetype == "parquet"
    assert info.schema.names == ["a", "b", "c"]
    assert info.rows == TABLE.num_rows
    assert info.size == (Path.cwd() / "data" / path).stat().st_size

    storage.delete(path)
    assert not storage.exists(path)


def test_info_arrow(storage: Storage) -> None:
    path = PurePath("foo/bar.arrow")

    storage.write(path, TABLE)
    assert storage.exists(path)

    info = storage.info(path)
    assert info is not None
    assert info.schema.names == ["a", "b", "c"]
    assert info.rows == TABLE.num_rows
    assert info.size == TABLE.nbytes

    storage.delete(path)
    assert not storage.exists(path)


def test_unhandled_extension(storage: Storage) -> None:
    path = PurePath("foo/bar.csv")

    # Read
    with pytest.raises(ValueError, match="Unknown file extension: .csv"):
        storage.read(path)

    # Write
    with pytest.raises(ValueError, match="Unknown file extension: .csv"):
        storage.write(path, TABLE)

    # Info
    with pytest.raises(ValueError, match="Unknown file extension: .csv"):
        storage.info(path)



from pathlib import PurePath

import pyarrow as pa

from renkon.repo import Registry
from renkon.repo.info import TableStat


def test_register_table(registry: Registry) -> None:
    store_table_info = TableStat(
        path=PurePath("tables/foo.arrow"),
        filetype="arrow",
        schema=pa.schema([pa.field("a", pa.int64()), pa.field("b", pa.string())]),
        rows=10,
        size=100,  # this is a made up number
    )

    path = PurePath("cells/13/df.arrow")
    registry.register("df[13]", path, store_table_info)

    reg_table_info = registry.lookup("df[13]", by="name")
    assert reg_table_info is not None
    assert reg_table_info.name == "df[13]"
    assert reg_table_info.path == path
    assert reg_table_info.schema == store_table_info.schema
    assert reg_table_info.rows == store_table_info.rows
    assert reg_table_info.size == store_table_info.size

    registry.unregister("df[13]")
    assert registry.lookup("df[13]", by="name") is None
    assert registry.lookup(str(path), by="path") is None


def test_list_tables(registry: Registry) -> None:
    named_table_infos = [
        (
            name,
            TableStat(
                path=PurePath(f"tables/{i}.{('arrow', 'parquet')[i % 2]}"),
                filetype=("arrow", "parquet")[i % 2],
                schema=pa.schema([pa.field("a", pa.int64()), pa.field("b", pa.string())]),
                rows=10,
                size=100,  # this is a made up number
            ),
        )
        for i, name in enumerate(["foo", "bar", "baz", "qux"])
    ]

    for name, table_info in named_table_infos:
        registry.register(name, PurePath(f"tables/{name}.{table_info.filetype}"), table_info)

    reg_table_infos = registry.list_all()
    assert len(reg_table_infos) == len(named_table_infos)
    for (name, table_info), reg_info in zip(named_table_infos, reg_table_infos, strict=True):
        assert reg_info.name == name
        assert reg_info.path == PurePath(f"tables/{name}.{table_info.filetype}")
        assert reg_info.schema == table_info.schema
        assert reg_info.rows == table_info.rows
        assert reg_info.size == table_info.size
